from logging import debug

import numpy as np

from ..objects.lq_hq_state import LQHQState
from ..objects.safe_rng import SafeRNG
from ..pipeline.utils import mix_seed


def node_iterator(nodes, state: LQHQState):
    for node in nodes:
        state = node(state)
    return state


def std_process(
    img_path: str, reader, saver, nodes, seed: int, out_path: str, image_dict: dict
):
    result_name = image_dict[img_path]
    debug(
        f"__________________________________\nImage path: {img_path} ResultName: {result_name}"
    )
    img = reader(img_path)
    state = LQHQState(rng=SafeRNG(mix_seed(seed, img_path)), hq=img, lq=img.copy())
    state = node_iterator(nodes, state)

    saver(state, out_path, result_name)


def tile_process(
    img_path: str,
    reader,
    saver,
    nodes,
    seed: int,
    out_path: str,
    image_dict: dict,
    wb: bool,
):
    for index, img in enumerate(reader(img_path)):
        if wb:
            mean = np.mean(img)
            if mean == 0.0 or mean == 1.0:
                continue
        out_name = f"{image_dict[img_path]}_{index}"
        debug(
            f"__________________________________\nImage path: {img_path} ResultName: {out_name} TileIndex: {index}"
        )
        state = LQHQState(
            rng=SafeRNG(mix_seed(seed, img_path, index)), hq=img, lq=img.copy()
        )
        state = nodes(state)
        saver(state, out_path, out_name)



import importlib.util
from . import core

if importlib.util.find_spec("pepedd.nodes"):
    from contextlib import suppress

    with suppress(ImportError):
        from . import nodes

        __all__ = ["core", "nodes"]
        # __all__ = ["core","*"]
else:
    __all__ = ["core"]



from datetime import datetime
import logging
from logging import debug
import os

import multiprocessing as mp
from concurrent.futures import ProcessPoolExecutor, as_completed
from functools import partial

from tqdm import tqdm
from tqdm.contrib.concurrent import thread_map

from pepeline import read, read_tiler, ImgFormat, ImgColor
from ..objects.safe_rng import SafeRNG
from ..node_register import registered_classes
from ..pipeline.process import std_process, tile_process
from ..pipeline.utils import (
    save_lq,
    save_hq_lq,
    out_clear,
    digits_count,
    get_best_context,
)
from ..pipeline.schema import PipelineOptions

from pathlib import Path


def init_worker(opts, img_names):
    global FORWARD
    from ..node_register import registered_classes

    nodes = [registered_classes[deg.type](deg.options) for deg in opts.degradation]
    if opts.tile:
        FORWARD = partial(
            tile_process,
            reader=partial(
                read_tiler,
                color_mode=ImgColor.GRAY if opts.gray else ImgColor.RGB,
                img_format=ImgFormat.F32,
                tile_size=opts.tile.size,
            ),
            saver=save_lq if opts.only_lq else save_hq_lq,
            nodes=nodes,
            seed=opts.seed,
            out_path=opts.output,
            image_dict=img_names,
            wb=opts.tile.no_wb,
        )

    else:
        FORWARD = partial(
            std_process,
            reader=partial(
                read,
                color_mode=ImgColor.GRAY if opts.gray else ImgColor.RGB,
                img_format=ImgFormat.F32,
            ),
            saver=save_lq if opts.only_lq else save_hq_lq,
            nodes=nodes,
            seed=opts.seed,
            out_path=opts.output,
            image_dict=img_names,
        )


def process_image(img_path):
    FORWARD(img_path)


class PipeLine:
    def __init__(self, options: PipelineOptions | dict):
        if isinstance(options, dict):
            opts = PipelineOptions(**options)
        else:
            opts = options
        if opts.debug:
            debug_folder = Path("debug")
            debug_folder.mkdir(exist_ok=True)

            timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
            log_file = debug_folder / f"{timestamp}.log"

            # Настраиваем логирование: пишем и в файл, и в консоль
            logging.basicConfig(
                level=logging.DEBUG,
                format="%(message)s",
                handlers=[
                    logging.FileHandler(log_file, encoding="utf-8"),
                    # logging.StreamHandler(),
                ],
            )
        self.in_path = opts.input
        img_names = os.listdir(self.in_path)
        if opts.dataset_size:
            img_names = img_names[: opts.dataset_size]
        if opts.shuffle_dataset:
            rng = SafeRNG(opts.seed)
            rng.shuffle(img_names)
        self.img_names = {}

        if opts.real_name:
            for index, img_name in enumerate(img_names):
                self.img_names[os.path.join(self.in_path, img_name)] = Path(
                    img_name
                ).stem
        else:
            z_fill = digits_count(len(img_names))
            for index, img_name in enumerate(img_names):
                self.img_names[os.path.join(self.in_path, img_name)] = str(index).zfill(
                    z_fill
                )

        self.out_path = opts.output
        self.num_workers = opts.num_workers
        self.map_type = opts.map_type

        if opts.output_clear:
            out_clear(self.out_path)
        os.makedirs(os.path.join(self.out_path, "lq"), exist_ok=True)
        if not opts.only_lq:
            os.makedirs(os.path.join(self.out_path, "hq"), exist_ok=True)
        self.opts = opts
        if not opts.map_type == "process":
            nodes = [
                registered_classes[deg.type](deg.options) for deg in opts.degradation
            ]
            if opts.tile:
                self.forward = partial(
                    tile_process,
                    reader=partial(
                        read_tiler,
                        color_mode=ImgColor.GRAY if opts.gray else ImgColor.RGB,
                        img_format=ImgFormat.F32,
                        tile_size=opts.tile.size,
                    ),
                    saver=save_lq if opts.only_lq else save_hq_lq,
                    nodes=nodes,
                    seed=opts.seed,
                    out_path=opts.output,
                    image_dict=self.img_names,
                    wb=opts.tile.no_wb,
                )

            else:
                self.forward = partial(
                    std_process,
                    reader=partial(
                        read,
                        color_mode=ImgColor.GRAY if opts.gray else ImgColor.RGB,
                        img_format=ImgFormat.F32,
                    ),
                    saver=save_lq if opts.only_lq else save_hq_lq,
                    nodes=nodes,
                    seed=opts.seed,
                    out_path=opts.output,
                    image_dict=self.img_names,
                )

        debug(f"Start\nSeed: {opts.seed}")

    def _run_simple(self):
        img_paths = list(self.img_names.keys())
        for img_path in tqdm(img_paths, desc="Processing images", unit="img"):
            self.forward(img_path)

    def _run_thread(self, num_workers=None):
        if num_workers is None:
            num_workers = mp.cpu_count()

        img_paths = list(self.img_names.keys())
        thread_map(
            self.forward,
            img_paths,
            max_workers=num_workers,
            desc="Processing images",
            unit="img",
        )

    def _run_process(self, num_workers=None):
        if num_workers is None:
            num_workers = mp.cpu_count()

        tasks = list(self.img_names.keys())
        ctx = get_best_context()

        with ProcessPoolExecutor(
            initializer=init_worker,
            initargs=(self.opts, self.img_names),
            max_workers=num_workers,
            mp_context=ctx,
        ) as executor:
            futures = {executor.submit(process_image, task): task for task in tasks}
            with tqdm(total=len(futures), desc="Processing images", unit="img") as pbar:
                for future in as_completed(futures):
                    try:
                        future.result()
                    except Exception as e:
                        print(f"\nError processing {futures[future]}: {e}")
                    pbar.update(1)

    def __call__(
        self,
    ):
        if self.map_type == "simple":
            print("Using simple sequential processing (for loop)")
            self._run_simple()
        elif self.map_type == "thread":
            print(
                f"Using thread-based parallelism ({self.num_workers or mp.cpu_count()} workers)"
            )
            self._run_thread(self.num_workers)
        elif self.map_type == "process":
            print(
                f"Using process-based parallelism ({self.num_workers or mp.cpu_count()} workers)"
            )
            self._run_process(self.num_workers)
        else:
            raise ValueError(
                f"Unknown map_type: {self.map_type}. Use 'simple', 'thread', or 'process'"
            )

        return



from numpy.random import default_rng


class SafeRNG:
    def __init__(self, seed: int = None):
        self.rng = default_rng(seed)

    def safe_uniform(self, rand_list: list[float] | float) -> float:
        if not isinstance(rand_list, list):
            return rand_list
        if len(rand_list) == 1 or rand_list[0] >= rand_list[1]:
            return rand_list[0]
        return self.rng.uniform(rand_list[0], rand_list[1])

    def safe_randint(self, rand_list: list[int] | int) -> int:
        if not isinstance(rand_list, list):
            return rand_list
        if len(rand_list) == 1 or rand_list[0] >= rand_list[1]:
            return rand_list[0]
        return int(self.rng.integers(rand_list[0], rand_list[1]))

    def __getattr__(self, name):
        return getattr(self.rng, name)



from abc import ABC, abstractmethod
from typing import Optional

from ..objects.lq_hq_state import LQHQState
from ..objects.safe_rng import SafeRNG


class Node(ABC):
    def __init__(self, probability: float = 1.0, seed: Optional[int] = None):
        self.seed = seed
        self.probability = probability

    def __call__(self, data: LQHQState) -> LQHQState:
        if self.seed is not None:
            rng_backup = data.rng
            data.rng = SafeRNG(seed=self.seed)
            if self.probability > data.rng.uniform(0, 1):
                data = self.forward(data)
            data.rng = rng_backup
        else:
            if self.probability > data.rng.uniform(0, 1):
                data = self.forward(data)
        return data

    @abstractmethod
    def forward(self, data: LQHQState) -> LQHQState:
        pass



registered_classes = {}


def register_class(key):
    """
    Decorator function to register a class with a given key.

    Args:
        key (str): The key to associate with the registered class.

    Returns:
        callable: Decorator function.

    """

    def decorator(cls):
        """
        Decorator function to register a class with a given key.

        Args:
            cls (class): The class to be registered.

        Returns:
            class: The input class.

        """
        registered_classes[key] = cls
        return cls

    return decorator



from .pipeline.pipeline import PipeLine, PipelineOptions
from .objects.lq_hq_state import LQHQState
from .objects.node_base import Node
from .objects.safe_rng import SafeRNG
from .pipeline.schema import Degradation


__all__ = [
    "PipeLine",
    "PipelineOptions",
    "LQHQState",
    "Node",
    "SafeRNG",
    "Degradation",
]



import hashlib
import math
import os
import shutil
import sys
import multiprocessing as mp
from ..objects.lq_hq_state import LQHQState
from pepeline import save


def save_lq(state: LQHQState, out_path: str, img_name: str):
    save(
        state.lq.clip(0, 1),
        os.path.join(out_path, "lq", f"{img_name}.png"),
    )


def save_hq_lq(state: LQHQState, out_path: str, img_name: str):
    save(
        state.lq.clip(0, 1),
        os.path.join(out_path, "lq", f"{img_name}.png"),
    )
    save(
        state.hq.clip(0, 1),
        os.path.join(out_path, "hq", f"{img_name}.png"),
    )


def digits_count(n: int) -> int:
    if n == 0:
        return 1
    return int(math.log10(abs(n))) + 1


def get_best_context():
    if sys.platform == "win32":
        return mp.get_context("spawn")

    try:
        return mp.get_context("forkserver")
    except ValueError:
        return mp.get_context("fork")


def mix_seed(*parts, bits=64) -> int:
    h = hashlib.blake2b(digest_size=bits // 8)
    for p in parts:
        if isinstance(p, bytes):
            h.update(p)
        else:
            h.update(str(p).encode("utf-8"))
        h.update(b"|")  # разделитель
    return int.from_bytes(h.digest(), "little")


def out_clear(folder):
    if os.path.exists(folder):
        shutil.rmtree(folder)



from typing import List, Any, Literal, Optional

from pydantic import BaseModel, model_validator


class Degradation(BaseModel):
    type: str
    options: Any


class TileOptions(BaseModel):
    size: int = 512
    no_wb: bool = True


class PipelineOptions(BaseModel):
    input: str = "input"
    output: str = "output"
    map_type: Literal["simple", "thread", "process"] = "thread"
    degradation: List[Degradation]

    num_workers: Optional[int] = None
    tile: Optional[TileOptions] = None

    dataset_size: Optional[int] = None
    shuffle_dataset: bool = True
    gray: bool = False

    debug: bool = False

    only_lq: bool = False
    real_name: bool = False

    output_clear: bool = True
    seed: Optional[int] = None

    @model_validator(mode="after")
    def apply_debug_and_seed(self):
        if self.debug:
            self.map_type = "simple"
            if self.seed is None:
                self.seed = 1234
        if self.seed is None:
            import secrets

            self.seed = secrets.randbits(64)
        return self
